/**
 * configurator.js
 * loads and configures classes and properties from SPARQL endpoint
 */

/**
 * first function to be called
 * queries and loads all classes from endpoint
 */
function loadClasses() {

    startTime = new Date();

    // try to use Sparkle if we have access to it
    if (typeof dataAgent != "undefined") {

        classList.clear();

        dataAgent.selectVisual(queryAllClasses, classConsumer);

        for (var i = 0; i < classList.size(); i++) {
            var javaClassInfo = classList.get(i);

            var fromJavaClass = {
                localName: javaClassInfo.getLocalName(),
                nameSpace: javaClassInfo.getNameSpace(),
                uri: javaClassInfo.getUri(),
                count: javaClassInfo.getCount(),
                id: i,
                attributes: [],
                objectProperties: [],
                selected: false,
                show: false,
                aggrCount: false,
                duplicates: [],
                duplicate: false,
                moreInfo: []
            };

            classes.push(fromJavaClass);

        }

        printInfo("Loading ontologies ...");

        loadOntologies(function () {

            printInfo("Loading prefixes, properties and additional info ...");

            updatePrefixes(function () {

                updateClassesWithPrefixes(function () {
                    assignInfoToClasses();
                    addPropertiesToClasses(function () {
                        constructGraph();
                        createClassSelection();
                        createAllAttributesSelection();

                        printSuccess("Application initialized");
                    });
                });
            });

        });

        // else load using connector
    } else {
        endpointCall(endpoint, queryAllClasses, function (data) {

            printInfo("Loading classes ...");

            var results = data.results.bindings;

            for (var i = 0; i < results.length; i++) {

                var javascriptClassInfo = createClassFromResult(results[i]);
                javascriptClassInfo.id = i;

                classes.push(javascriptClassInfo);

            }

            printInfo("Loading ontologies ...");

            loadOntologies(function () {

                printInfo("Loading prefixes, properties and additional info ...");

                updatePrefixes(function () {

                    updateClassesWithPrefixes(function () {
                        assignInfoToClasses();
                        addPropertiesToClasses(function () {
                            constructGraph();
                            createClassSelection();
                            createAllAttributesSelection();

                            printSuccess("Application initialized");
                        });
                    });
                });

            });


        });
    }
}

/**
 * assigns a type of filter (data type) to an attribute based on its range
 * @param attribute
 */
function assignFilterTypeTo(attribute) {

    var range = getRange(attribute);

    var datatype = getDatatypeForRange(range);

    if (!datatype) {
        attribute.filterType = "string";
    } else {
        attribute.filterType = datatype;
    }

    createFilterSelection(attribute);
    createAggregationsSelection(attribute);

}

/**
 * loads and assigns properties to their classes
 * @param callback
 */
function addPropertiesToClasses(callback) {

    var promises = [];
    var propertiesArray = [];

    // asynchronous operation, need to wait for them with promises
    for (var i = 0; i < classes.length; i++) {

        var query = "SELECT DISTINCT ?property (count(?x) AS ?cnt) WHERE { ?x a <" + classes[i].uri + "> .  ?x ?property ?o . } GROUP BY ?property ORDER BY DESC(?cnt)";

        var promise = callAndGetProperties(query, propertiesArray, i);
        promises.push(promise);

    }

    // once all properties are loaded
    Promise.all(promises).then(function () {


        for (var j = 0; j < propertiesArray.length; j++) {

            var properties = propertiesArray[j];

            // assign all properties to thier class
            for (var k = 0; k < properties.length; k++) {


                var property = properties[k];

                var prefix = getPrefixByNameSpace(property.nameSpace);

                var prefixedLocalName = prefix + ":" + property.localName;

                var newAttribute = {
                    count: property.count,
                    localName: property.localName,
                    selected: property.selected,
                    nameSpace: property.nameSpace,
                    uri: property.uri,
                    prefixedLocalName: prefixedLocalName,
                    prefix: prefix,
                    show: property.show,
                    optional: property.optional,
                    filters: property.filters,
                    aggregations: property.aggregations,
                    moreInfo: []
                };

                // check if the property is an annotation
                if (arrayIncludes(annotation, newAttribute.uri)) {
                    newAttribute.annotation = true;
                }

                // loading more information about the property from respective ontology

                var xml = getOntologyByNameSpace(property.nameSpace);

                if (!xml) {
                    classes[j].attributes.push(newAttribute);
                    continue;
                }

                var element = getElementByAttributeValue(xml, "rdf:about", property.uri);

                if (!element) {
                    continue;
                }

                newAttribute.type = element.localName;

                for (var o = 0; o < element.childNodes.length; o++) {

                    var child = element.childNodes[o];

                    // load domains
                    if (child.localName === "domain") {

                        if (typeof newAttribute.domain == "undefined") {
                            newAttribute.domain = [];
                        }

                        if (child.attributes.length === 0) {
                            continue;
                        }

                        newAttribute.domain.push(child.attributes[0].value);

                        // load ranges
                    } else if (child.localName === "range") {

                        if (typeof newAttribute.range == "undefined") {
                            newAttribute.range = [];
                        }
                        newAttribute.range.push(child.attributes[0].value);

                        // load annotations
                    } else if (child.localName === "comment" || child.localName === "label" || child.localName === "title" || child.localName === "prefLabel") {
                        newAttribute.moreInfo.push(createAnnotation(child));
                    }
                }


                // divide properties to object properties and attributes
                if (newAttribute.type === "ObjectProperty") {

                    if (typeof newAttribute.range != "undefined" && newAttribute.range.length > 1) {
                        for (var i = 0; i < newAttribute.range.length; i++) {
                            newAttribute.selectedRange = i;
                            newAttribute.selectedRange = newAttribute.range[i];

                            classes[j].objectProperties.push(newAttribute);
                        }
                    } else {
                        classes[j].objectProperties.push(newAttribute);
                    }


                } else {
                    classes[j].attributes.push(newAttribute);
                }

            }
        }


        callback();

    });


}

/**
 * queries endpoint and gets properties of their respective classes
 * asynchronous
 * @param query the query to get properties
 * @param mainPropertiesArray the array of arrays of properties
 * @param index index of a class
 * @returns {Promise<any>}
 */
function callAndGetProperties(query, mainPropertiesArray, index) {

    var propertiesArray = [];

    var resolve, reject;

    var promise = new Promise(function (_resolve, _reject) {
        resolve = _resolve;
        reject = _reject;
    });

    // try to use Sparkle
    if (typeof dataAgent != "undefined") {

        classList.clear();

        dataAgent.selectVisual(query, classConsumer);

        for (var i = 0; i < classList.size(); i++) {
            var javaClassInfo = classList.get(i);

            var javasClassInfo = {
                localName: javaClassInfo.getLocalName(),
                nameSpace: javaClassInfo.getNameSpace(),
                uri: javaClassInfo.getUri(),
                count: javaClassInfo.getCount(),
                selected: false,
                optional: true,
                show: true,
                aggregations: [],
                filters: [],
                clicked: false
            };

            propertiesArray.push(javasClassInfo);

        }

        mainPropertiesArray[index] = propertiesArray;
        resolve(propertiesArray);

        // no Sparkle
    } else {
        endpointCall(endpoint, query, function (data) {

            var results = data.results.bindings;

            for (var i = 0; i < results.length; i++) {

                var javascriptClassInfo = createPropertyFromResult(results[i]);

                propertiesArray.push(javascriptClassInfo);

            }

            // found properties pushed to the main properties array based on its class index
            mainPropertiesArray[index] = propertiesArray;
            resolve(propertiesArray);

        });
    }


    return promise;

}

/**
 * tries to load information about classes from their ontologies and assigns it to them
 */
function assignInfoToClasses() {
    for (var i = 0; i < classes.length; i++) {

        var cls = classes[i];

        var xml = getOntologyByNameSpace(cls.nameSpace);

        if (!xml) {
            continue;
        }

        var element = getElementByAttributeValue(xml, "rdf:about", cls.uri);

        if (!element) {
            continue;
        }

        for (var k = 0; k < element.childNodes.length; k++) {
            var info = element.childNodes[k];

            // subclasses
            if (info.localName === "subClassOf") {
                // TODO female in dasta.owl collection
                if (info.attributes.length > 0) {
                    cls.subClassOf = info.attributes[0].value;
                }

                // annotations
            } else if (info.localName === "comment" || info.localName === "label" || info.localName === "title" || info.localName === "prefLabel") {
                cls.moreInfo.push(createAnnotation(info));
            }

        }
    }

}

/**
 * supporting function to create an object of annotation for moreInfo array of a class
 * @param element
 * @returns {{name: string, tag: string, value: (string|string)}}
 */
function createAnnotation(element) {

    var name = element.localName;
    var tag;

    if (element.attributes.length > 0) {
        tag = element.attributes[0].value;
    } else {
        tag = "";
    }

    var value = element.textContent;

    return {
        name: name,
        tag: tag,
        value: value
    };

}

/**
 * updates classes with their prefixes - prefix and prefixedLocalName attributes
 * finds their prefix by the class namespace
 * @param callback
 */
function updateClassesWithPrefixes(callback) {

    for (var i = 0; i < classes.length; i++) {

        var cls = classes[i];

        var prefix = getPrefixByNameSpace(cls.nameSpace);

        cls.prefixedLocalName = prefix + ":" + cls.localName;
        cls.prefix = prefix;

    }

    callback();

}

/**
 * tries to create prefixes from ontologies - all prefixes defined in the ontologies are loaded
 * if application needs a prefix of a namespace which won't be defined here, it will use the default one instead
 * @param callback
 */
function updatePrefixes(callback) {

    for (var i = 0; i < ontologies.length; i++) {

        if (ontologies[i].xml == null) {
            continue;
        }

        var localPrefixes = ontologies[i].xml.documentElement.attributes;


        for (var j = 0; j < localPrefixes.length; j++) {

            var prefix = localPrefixes[j];

            // skip useless prefixes
            if (prefix.localName === "xmlns" || prefix.localName === "base") {
                continue;
            }

            // check if prefix al already present
            if (checkForPrefix(prefix.value) === true) {
                continue;
            }

            var uriClass;

            if (prefix.value.indexOf("#") > -1) {
                uriClass = prefix.value.substring(0, prefix.value.length - 1);
            } else {
                uriClass = prefix.value;
            }

            var newPrefix = {
                prefix: prefix.localName,
                uri: prefix.value,
                uriClass: uriClass,
                uriSPARQL: "<" + prefix.value + ">"
            };


            prefixes.push(newPrefix);

        }
    }

    callback();
}

