/**
 * connector.js
 * contains functions for loading of ontologies and querying SPARQL endpoint
 */

/**
 * loads an ontology of given url
 * if using Sparkle, loads it from the application
 * returns XML representation of loaded ontology
 * @param url
 * @param callback
 */
function ontologyCall(url, callback) {

    if (typeof ontologyLoader != "undefined") {

        var response = ontologyLoader.getOntology(url);

        var parser = new DOMParser();

        callback(parser.parseFromString(response, "text/xml"));

    } else {

        var address;

        if (typeof cors !== "undefined") {
            address = cors + url;
        } else {
            address = url;
        }

        $.get(address, function (data, status) {
            callback(data);
        });
    }
}

/**
 * queries an endpoint using its address and query
 * returns the endpoint response
 * @param sparqlEndpoint
 * @param query
 * @param callback
 */
function endpointCall(sparqlEndpoint, query, callback) {
    var queryUrl = sparqlEndpoint + "?query=" + encodeURIComponent(query) + "&format=json";

    $.get(queryUrl, function (data, status) {
        callback(data);
    });
}

/**
 * function to load all ontologies specified in options
 * syncrhonous, since it calls callback only when done
 * @param callback
 */
function loadOntologies(callback) {

    var promises = [];

    for (var i = 0; i < ontologies.length; i++) {

        var promise = loadOntology(ontologies[i]);

        promises.push(promise);

    }

    Promise.all(promises).then(function () {
        callback();
    });
}

/**
 * asynchronous function, returns a XML representation of given ontology
 * @param ontology
 * @returns {Promise<any>}
 */
function loadOntology(ontology) {

    var resolve, reject;

    var promise = new Promise(function (_resolve, _reject) {
        resolve = _resolve;
        reject = _reject;
    });

    ontologyCall(ontology.url, function (response) {

        if (typeof response === "undefined") {
            resolve(ontology);
        }

        ontology.xml = response;

        resolve(ontology);
    });

    return promise;
}

/**
 * supporting function to create a class object from an ednpoint result
 * @param result
 * @returns {{localName: string, duplicates: Array, count: *, show: boolean, attributes: Array, objectProperties: Array, nameSpace: string, aggrCount: boolean, duplicate: boolean, uri: *, moreInfo: Array, selected: boolean}}
 */
function createClassFromResult(result) {
    var uri = result.class.value;

    var nameSpace, localName;

    if (uri.indexOf("#") > -1) {
        var split = uri.split("#");
        nameSpace = split[0];
        localName = split[1];
    } else {
        nameSpace = uri.substring(0, uri.lastIndexOf("/") + 1);
        localName = uri.substring(uri.lastIndexOf("/") + 1, uri.length);
    }

    var count = result.cnt.value;

    return {
        localName: localName,
        nameSpace: nameSpace,
        uri: uri,
        count: count,
        attributes: [],
        objectProperties: [],
        selected: false,
        show: false,
        aggrCount: false,
        duplicates: [],
        duplicate: false,
        moreInfo: []
    };
}

/**
 * supporting function to create a property object from an endpoint result
 * @param result
 * @returns {{localName: string, count: *, show: boolean, optional: boolean, nameSpace: string, filters: Array, uri: *, aggregations: Array, clicked: boolean, selected: boolean}}
 */
function createPropertyFromResult(result) {
    var uri = result.property.value;

    var nameSpace, localName;

    if (uri.indexOf("#") > -1) {
        var split = uri.split("#");
        nameSpace = split[0];
        localName = split[1];
    } else {
        nameSpace = uri.substring(0, uri.lastIndexOf("/") + 1);
        localName = uri.substring(uri.lastIndexOf("/") + 1, uri.length);
    }

    var count = result.cnt.value;

    return {
        localName: localName,
        nameSpace: nameSpace,
        uri: uri,
        count: count,
        selected: false,
        optional: true,
        show: true,
        aggregations: [],
        filters: [],
        clicked: false
    };
}