/**
 * helper.js
 * supporting functions for application
 */

/**
 * checks whether given variable is present in orderBy array (is ordered)
 * @param variable
 * @returns {*}
 */
function isVariableOrdered(variable) {

    for (var i = 0; i < orderBy.length; i++) {
        if (orderBy[i].variable === variable) {
            return i;
        }
    }

    return false;

}

/**
 * checks selected edges of a class and assigns the same to its mirror
 * @param clsId
 * @param mirrorId
 */
function addExistingEdgesToViz(clsId, mirrorId) {

    for (var i = 0; i < classes.length; i++) {

        if (i === mirrorId) {
            continue;
        }

        if (i === clsId) {

            for (var o = 0; o < classes[i].objectProperties.length; o++) {

                if (classes[i].objectProperties[o].selected === true) {

                    var newEdge = {
                        source: classes[mirrorId].d3Id,
                        target: classes[classes[i].objectProperties[o].target].d3Id,
                        name: classes[i].objectProperties[o].prefixedLocalName,
                        edgeName: classes[i].objectProperties[o].localName
                    };
                    addEdge(newEdge);
                }

            }

        }

        for (var j = 0; j < classes[i].objectProperties.length; j++) {

            if (classes[i].objectProperties[j].selected === true && classes[i].objectProperties[j].target === mirrorId) {

                var nwEdge = {
                    source: classes[i].d3Id,
                    target: classes[mirrorId].d3Id,
                    name: classes[i].objectProperties[j].prefixedLocalName,
                    edgeName: classes[i].objectProperties[j].localName
                };

                addEdge(nwEdge);

            }

        }

    }

}

/**
 * creates a duplicate object class to given class - mirroring
 * attributes are unselected
 * @param cls
 */
function createDuplicateClass(cls) {

    var duplicate = createDuplicate(cls);

    duplicate.attributes = [];
    duplicate.objectProperties = [];

    for (var i = 0; i < cls.attributes.length; i++) {

        var attr = createDuplicate(cls.attributes[i]);

        attr.filters = [];
        attr.aggregations = [];

        attr.selected = false;

        duplicate.attributes.push(attr);

    }

    for (var j = 0; j < cls.objectProperties.length; j++) {

        var objectProperty = createDuplicate(cls.objectProperties[j]);


        duplicate.objectProperties.push(objectProperty);

    }

    return duplicate;

}

/**
 * creates a copy of an object
 * @param object
 */
function createDuplicate(object) {
    var duplicate = {};

    for (var u in object) {
        duplicate[u] = object[u];
    }

    return duplicate;
}

/**
 *
 * @param str
 * @param find
 * @param replace
 * @returns {void | string | never}
 */
function replaceAll(str, find, replace) {
    return str.replace(new RegExp(escapeRegExp(find), 'g'), replace);
}

/**
 *
 * @param str
 * @returns {void | string | never}
 */
function escapeRegExp(str) {
    return str.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1");
}


/**
 * returns a datatype for given range (string, number, date)
 * @param range
 * @returns {*}
 */
function getDatatypeForRange(range) {
    for (var i = 0; i < datatypes.length; i++) {
        if (range === datatypes[i].uri) {
            return datatypes[i].type;
        }
    }

    return false;
}

/**
 * returns if a prefix is defined for given uri
 * @param uri
 * @returns {boolean}
 */
function checkForPrefix(uri) {

    for (var i = 0; i < prefixes.length; i++) {
        if (prefixes[i].uri === uri) {
            return true;
        }
    }


    return false;
}

/**
 * returns ontology xml file for given namespace
 * @param nameSpace
 * @returns {*}
 */
function getOntologyByNameSpace(nameSpace) {

    var nmsp;

    if (nameSpace.indexOf("#") > -1) {
        nmsp = nameSpace.substring(0, nameSpace.length - 1);
    } else {
        nmsp = nameSpace;
    }

    for (var i = 0; i < ontologies.length; i++) {
        if (ontologies[i].nameSpace === nmsp) {
            return ontologies[i].xml;
        }
    }

    return false;
}

/**
 * returns an element with given attribute and its value (like rdf:about)
 * @param xml
 * @param attr
 * @param attrValue
 * @returns {*}
 */
function getElementByAttributeValue(xml, attr, attrValue) {

    var elements = xml.documentElement.childNodes;

    for (var i = 0; i < elements.length; i++) {
        var elem = elements[i];

        if (typeof elem.attributes === "undefined") {
            continue;
        }

        if (elem.attributes.length > 0) {

            if (elem.attributes[0].nodeName === attr && elem.attributes[0].value === attrValue) {
                return elem;
            }
        }
    }

    return false;


}

/**
 * returns prefix object based on the prefix (because we use only arrays -> possibility of set here)
 * @param prefix
 * @returns {*}
 */
function getPrefixByPrefix(prefix) {
    for (var i = 0; i < prefixes.length; i++) {

        if (prefixes[i].prefix === prefix) {
            return prefixes[i];
        }

    }

    return false;
}

/**
 * returns a prefix for given namespace
 * if prefix does not exist it creates a new one
 * @param nameSpace
 * @returns {*}
 */
function getPrefixByNameSpace(nameSpace) {

    for (var i = 0; i < prefixes.length; i++) {

        if (prefixes[i].uriClass === nameSpace || prefixes[i].uri === nameSpace) {
            return prefixes[i].prefix;
        }
    }

    // prefix not found
    var newPrefix = {
        prefix: "prx" + prefixIndex++,
        uri: nameSpace,
        uriClass: nameSpace,
        uriSPARQL: "<" + nameSpace + ">"
    };

    prefixes.push(newPrefix);

    return newPrefix.prefix;

}

/**
 * returns an array of classes which have given subclass as their subClassOf property
 * @param subClass
 * @returns {Array}
 */
function getAllClassesOfSubclass(subClass) {

    var localClasses = [];

    for (var i = 0; i < classes.length; i++) {

        if (typeof classes[i].subClassOf != "undefined") {

            if (classes[i].subClassOf === subClass) {
                localClasses.push(i);
            }

        }

    }

    return localClasses;


}

/**
 * returns range of given attribute
 * returns -1 if no range is present
 * @param property
 * @returns {*}
 */
function getRange(property) {

    if (typeof property.selectedRange != "undefined") {
        return property.selectedRange;
    }

    if (typeof property.range != "undefined") {

        if (property.range.length > 0) {
            return property.range[0];
        }

    }

    return -1;

}

/**
 *
 * @param prefixedLocalName
 * @returns {boolean|*}
 */
function getClassByPrefixedLocalName(prefixedLocalName) {
    for (var i = 0; i < classes.length; i++) {
        if (classes[i].prefixedLocalName === prefixedLocalName) {
            return classes[i];
        }
    }

    return false;
}

/**
 *  returns class object based on range of attribute
 * @param property
 * @returns {*}
 */
function getTargetFromProperty(property) {

    if (typeof property.selectedRange != "undefined") {
        return getClassIdByUri(property.selectedRange);
    }

    if (typeof property.range != "undefined") {

        if (property.range.length > 0) {
            return getClassIdByUri(property.range[0]);
        }

    }

    return -1;


}

/**
 * returns a class index based on its uri
 * @param uri
 * @returns {number}
 */
function getClassIdByUri(uri) {

    for (var i = 0; i < classes.length; i++) {

        if (classes[i].uri === uri) {
            return i;
        }

        if (typeof classes[i].subClassOf != "undefined") {
            if (classes[i].subClassOf === uri) {
                subClass = true;
                return i;
            }
        }
    }

    return -1;


}

/**
 * returns if given array includes given element (Array.includes())
 * @param array
 * @param what
 * @returns {boolean}
 */
function arrayIncludes(array, what) {
    for (var i = 0; i < array.length; i++) {
        if (array[i] === what) {
            return true;
        }
    }

    return false;
}


/**
 * returns an object property based on given source and target id
 * @param source
 * @param target
 * @param localName
 * @returns {*}
 */
function getObjectPropertyBySourceAndTarget(source, target, localName) {

    var cls = classes[source];

    for (var i = 0; i < cls.objectProperties.length; i++) {

        if (localName) {
            if (cls.objectProperties[i].target === target && cls.objectProperties[i].localName === localName) {
                return cls.objectProperties[i];
            }
        } else {
            if (cls.objectProperties[i].target === target) {
                return cls.objectProperties[i];
            }
        }

    }

    return false;

}

/**
 * compares arrays by their lengths
 * @param a
 * @param b
 * @returns {number}
 */
function compareByArrayLength(a, b) {
    if (a.length < b.length)
        return -1;
    if (a.length > b.length)
        return 1;
    return 0;
}

/**
 * compares elements by their count
 * @param a
 * @param b
 * @returns {number}
 */
function compareCount(a, b) {
    if (a.count < b.count) {
        return 1;
    }
    if (a.count > b.count) {
        return -1;
    }
    return 0;
}

/**
 * returns whether given node is selected (present in d3 visualization)
 * @param node
 * @returns {boolean}
 */
function isNodeSelected(node) {
    for (var i = 0; i < dataset.nodes.length; i++) {
        if (dataset.nodes[i].id === node) {
            return true;
        }
    }

    return false;
}