/**
 * query.js
 * contains functions for query generation
 */

/**
 * generated query without LIMIT and OFFSET
 * @type {string}
 */
var notLimitedQuery = "";

/**
 * runs generated query
 * if these is none, does not do anything
 * automatically creates a result table
 */
function runQuery() {
    if (printedQuery === "") {
        return;
    }


    endpointCall(endpoint, printedQuery, function (results) {

        queryResults = results;

        createResultsTable();

    });
}


function runCSV() {

    if (printedQuery === "") {
        return;
    }

    endpointCall(endpoint, printedQuery, function (results) {

        queryResults = results;

        var csv = createResultsTable();

        var csvFile = new Blob([csv], {type: "text/csv"});
        var downloadLink = document.createElement("a");
        downloadLink.download = "results.csv";
        downloadLink.href = window.URL.createObjectURL(csvFile);
        downloadLink.style.display = "none";
        document.body.appendChild(downloadLink);
        downloadLink.click();

    });

}

/**
 * returns a list of all variables in the query
 * @returns {Array}
 */
function getVariables() {
    var orderByVariables = [];

    for (var v = 0; v < query.variables.length; v++) {
        orderByVariables.push(query.variables[v]);
    }

    for (var a = 0; a < query.aggregations.length; a++) {
        orderByVariables.push(query.aggregations[a].variable);
    }


    return orderByVariables;
}

/**
 * function generates SPARQL query
 */
function generateQuery() {

    var selectedNodes = [];

    // prepare selected nodes
    for (var i = 0; i < classes.length; i++) {

        if (classes[i].selected === true) {
            selectedNodes.push(i);
        }
    }

    // null previous attemts
    query.prefixes = [];
    query.variables = [];
    query.where = [];
    query.filters = [];
    query.aggregations = [];

    // if there are no selected nodes -> we are done
    if (selectedNodes.length === 0) {
        document.getElementById("generatedQuery").innerHTML = "";
        createOrderBySelection(getVariables());
        createOrderByTable();
        return;
    }

    // find a starting node
    var startingNode = findStartingNode(selectedNodes);


    // init and go
    var stack = [startingNode];
    var visited = [];

    DFS(query, stack, visited);

    // a check if we have been in every selected node
    var check = checkAllNodesWereVisited(selectedNodes, visited);

    // while there are nodes we have not been to, go DFS again
    while (check !== true) {

        var newStart = findStartingNode(check);

        stack.push(newStart);
        DFS(query, stack, visited);
        check = checkAllNodesWereVisited(selectedNodes, visited);
    }

    createOrderBySelection(getVariables());
    createOrderByTable();


    var qr = printQuery(query);

    printedQuery = qr;

    tryQuery(qr);

    // if Sparkle, save the query for other editors
    if (typeof querySaver != "undefined") {
        querySaver.setQuery(qr);
    }
}

/**
 * function tries to run query against SPARQL endpoint with LIMIT 5
 * as a result, user can see if the query returned any results
 */
function tryQuery() {

    if (notLimitedQuery === "") {
        document.getElementById("query-result").innerHTML = "";
        return;
    }

    var q = notLimitedQuery + "\nLIMIT 5";

    // using Sparkle
    if (typeof dataAgent != "undefined") {

        classList.clear();

        dataAgent.selectVisual(q, classConsumer);

        if (classList.size() > 0) {
            document.getElementById("query-result").innerHTML = "Query returns results";
        } else {
            document.getElementById("query-result").innerHTML = "Query does not return any results!";
        }

        // or not
    } else {
        endpointCall(endpoint, q, function (data) {

            if (data.results === "undefined") {
                return;
            }

            var results = data.results.bindings;

            if (results.length === 0) {
                document.getElementById("query-result").innerHTML = "Query does not return any results!";
            } else {
                document.getElementById("query-result").innerHTML = "Query returns results";
            }
        });
    }


}

/**
 * checks whether all nodes where visited
 * returns true if all visited and list of not visited if any
 * @param selectedNodes
 * @param visitedNodes
 * @returns {boolean|Array}
 */
function checkAllNodesWereVisited(selectedNodes, visitedNodes) {

    var notSelected = [];

    for (var i = 0; i < selectedNodes.length; i++) {

        if (!arrayIncludes(visitedNodes, selectedNodes[i])) {
            notSelected.push(selectedNodes[i]);
        }
    }

    if (notSelected.length > 0) {
        return notSelected;
    } else {
        return true;
    }


}

/**
 * DFS search of selected nodes
 * generates query while traversing selected graph
 * @param query query object
 * @param stack nodes to visit
 * @param visited visited nodes
 */
function DFS(query, stack, visited) {

    while (stack.length > 0) {

        var start = stack.pop();

        // add to variables the class if set to show and it is not already in variables
        if (classes[start].show === true && !arrayIncludes(query.variables, classes[start].localName)) {
            query.variables.push(classes[start].localName);
        }

        // add COUNT aggregation if selected
        if (classes[start].aggrCount === true) {
            query.aggregations.push({
                aggr: "COUNT",
                attribute: classes[start].localName,
                variable: "count_" + classes[start].localName
            });
        }

        // handle class attributes
        var selectedAttributes = getSelectedAttributes(start);

        for (var s = 0; s < selectedAttributes.length; s++) {

            var attrVariable = classes[start].localName + "_" + selectedAttributes[s].localName;

            // add prefix to variables if not already present
            if (!arrayIncludes(query.prefixes, selectedAttributes[s].prefix)) {
                query.prefixes.push(selectedAttributes[s].prefix);
            }

            // add to variables only if show
            if (selectedAttributes[s].show === true) {
                query.variables.push(attrVariable);
            }

            // always add the attribute triple
            query.where.push({
                subject: classes[start].localName,
                predicate: selectedAttributes[s].prefixedLocalName,
                object: attrVariable,
                optional: selectedAttributes[s].optional
            });

            // handle filters next
            for (var f = 0; f < selectedAttributes[s].filters.length; f++) {
                var condition = selectedAttributes[s].filters[f];

                if (condition.value === "") {
                    continue;
                }

                var range = getRange(selectedAttributes[s]);

                // HACK - default string?
                if (range === -1) {
                    range = "http://www.w3.org/2001/XMLSchema#string";
                }

                var nameSpaceOfRange, localDataType;

                if (range.indexOf("#") > -1) {
                    var split = range.split("#");
                    nameSpaceOfRange = split[0];
                    localDataType = split[1];
                } else {
                    nameSpaceOfRange = range.substring(0, range.lastIndexOf("/") + 1);
                    localDataType = range.substring(range.lastIndexOf("/") + 1, range.length);
                }

                var prefix = getPrefixByNameSpace(nameSpaceOfRange);

                // add prefix if needed
                if (!arrayIncludes(query.prefixes, prefix)) {
                    query.prefixes.push(prefix);
                }

                // add the filter object
                query.filters.push({
                    type: selectedAttributes[s].filterType,
                    operator: condition.options[condition.selectedOption],
                    value: condition.value,
                    attribute: attrVariable,
                    dataType: prefix + ":" + localDataType
                });

            }

            // handle aggregations last
            for (var a = 0; a < selectedAttributes[s].aggregations.length; a++) {

                var aggregation = selectedAttributes[s].aggregations[a];

                // only if selected
                if (aggregation.selected === false) {
                    continue;
                }

                var aggrAttr = classes[start].localName + "_" + aggregation.attribute;

                // aggregation object
                query.aggregations.push({
                    aggr: aggregation.aggr,
                    attribute: aggrAttr,
                    variable: aggregation.variable
                });
            }
        }

        // find neighbours
        var neighbours = getSelectedNeighbours(start);

        for (var j = 0; j < neighbours.length; j++) {

            var next = neighbours[j];

            // add the neighbour to stack if not visited yet
            if (arrayIncludes(visited, next.target)) {
                continue;
            }

            if (!arrayIncludes(stack, next.target)) {

                // relation triple
                query.where.push({
                    subject: classes[start].localName,
                    predicate: next.prefixedLocalName,
                    object: classes[next.target].localName
                });

                // add to stack!
                stack.push(next.target);

                if (!arrayIncludes(query.prefixes, next.prefix)) {
                    query.prefixes.push(next.prefix);
                }


                var objectVariable = classes[next.target].localName;
                var prfx = classes[next.target].prefix;

                // add to variables if not present and show
                if (classes[next.target].show === true && !arrayIncludes(query.variables, objectVariable)) {
                    query.variables.push(objectVariable);
                }

                var subjectVariable = classes[start].localName;

                // check if our class is there and should be
                if (classes[start].show === true && !arrayIncludes(query.variables, subjectVariable)) {
                    query.variables.push(subjectVariable);
                }

                // don't forget prefix
                if (!arrayIncludes(query.prefixes, prfx)) {
                    query.prefixes.push(prfx);
                }
            }

        }

        visited.push(start);
    }
}

/**
 * function prints generated query in a readable way
 * @param query generated query
 * @returns {string}
 */
function printQuery(query) {

    var printedQuery = "";

    document.getElementById("generatedQuery").innerHTML = printedQuery;

    var prefixes = query.prefixes;
    var variables = query.variables;
    var filters = query.filters;
    var aggregations = query.aggregations;
    var conditions = query.where;

    // if there are no variables, the query is pointless
    if (variables.length === 0 && aggregations.length === 0) {
        return printedQuery;
    }

    // PREFIX
    for (var p = 0; p < prefixes.length; p++) {

        var prefix = getPrefixByPrefix(prefixes[p]);
        printedQuery += "PREFIX " + prefix.prefix + ": <" + prefix.uri + ">\n";

    }

    for (var z = 0; z < orderBy.length; z++) {
        if (orderBy[z].selectedOrderOption > 1) {
            if (!arrayIncludes(prefixes, "xsd")) {
                printedQuery += "PREFIX xsd: <" + getPrefixByPrefix("xsd").uri + ">\n";
                break;
            }
        }
    }

    // SELECT
    printedQuery += "\nSELECT ";

    if (query.DISTINCT === true) {
        printedQuery += "DISTINCT ";
    }

    // selected variables
    for (var v = 0; v < variables.length; v++) {

        printedQuery += "?" + variables[v] + " ";

    }

    if (aggregations.length > 0) {
        printedQuery += "\n";
    }

    // selected aggregations
    for (var a = 0; a < aggregations.length; a++) {
        printedQuery += "(" + aggregations[a].aggr + "(?" + aggregations[a].attribute + ") AS ?" + aggregations[a].variable + ") ";
    }

    // WHERE
    printedQuery += "\nWHERE {\n";


    // conditions
    for (var c = 0; c < conditions.length; c++) {

        var condition = printCondition(conditions[c]);

        printedQuery += condition + "\n";

    }

    // filters
    for (var f = 0; f < filters.length; f++) {

        var filter = printFilter(filters[f]);
        printedQuery += filter + "\n";

    }


    printedQuery += "}";

    // GROUP BY
    if (aggregations.length > 0 && variables.length > 0) {

        printedQuery += "\nGROUP BY ";

        for (var aj = 0; aj < variables.length; aj++) {

            printedQuery += "?" + variables[aj] + " ";


        }
    }


    // ORDER BY
    if (orderBy.length > 0) {
        printedQuery += "\nORDER BY ";
    }

    for (var o = 0; o < orderBy.length; o++) {

        var order = orderBy[o];

        var mod = order.selectedOrderOption % 2 === 0 ? "ASC" : "DESC";

        var orderedOption = order.orderOptions[order.selectedOrderOption];

        if (orderedOption.indexOf("int") > 1) {
            printedQuery += mod + "(xsd:int(?" + order.variable + ")) ";
        } else if (orderedOption.indexOf("datetime") > 1) {
            printedQuery += mod + "(xsd:datetime(?" + order.variable + ")) ";
        } else if (orderedOption.indexOf("date") > 1) {
            printedQuery += mod + "(xsd:date(?" + order.variable + ")) ";
        } else if (orderedOption.indexOf("time") > 1) {
            printedQuery += mod + "(xsd:time(?" + order.variable + ")) ";
        } else if (orderedOption.indexOf("float") > 1) {
            printedQuery += mod + "(xsd:float(?" + order.variable + ")) ";
        } else {
            printedQuery += mod + "(?" + order.variable + ") ";
        }
    }

    notLimitedQuery = printedQuery;

    // OFFSET
    if (query.OFFSET > 0) {
        printedQuery += "\nOFFSET " + query.OFFSET;
    }

    // LIMIT
    if (query.LIMIT > 0) {
        printedQuery += "\nLIMIT " + query.LIMIT;
    }


    // since we use <pre> tag, we need to encode "<" and ">" and print that
    var uncodedQuery = printedQuery;


    uncodedQuery = replaceAll(uncodedQuery, ">", "&gt;");
    uncodedQuery = replaceAll(uncodedQuery, "<", "&lt;");

    document.getElementById("generatedQuery").innerHTML = uncodedQuery;

    // returned is actually the not coded query
    return printedQuery;

}

function isAggregate(aggregations, variable) {
    for (var i = 0; i < aggregations.length; i++) {
        if (aggregations[i].attribute === variable) {
            return true;
        }
    }

    return false;
}

/**
 * returns a string representation of a condition
 * @param condition
 * @returns {string}
 */
function printCondition(condition) {

    var print = "";

    if (condition.optional === true) {
        print += "OPTIONAL { ";
    }

    print += "?" + condition.subject + " " + condition.predicate + " ?" + condition.object;

    if (condition.optional === true) {
        print += " }";
    } else {
        print += " . ";
    }

    return print;
}

/**
 * returns a string representation of a filter
 * @param filter
 * @returns {string}
 */
function printFilter(filter) {

    var value;


    if (filter.operator === "contains") {
        value = filter.value;
        return "FILTER ((regex(str(?" + filter.attribute + "), '" + value + "', 'i'))) . ";
    } else if (filter.operator === "is equal to") {
        value = "^" + filter.value + "$";
        return "FILTER ((regex(str(?" + filter.attribute + "), '" + value + "', 'i'))) . ";
    } else if (filter.operator === "not equal to") {
        value = "^" + filter.value + "$";
        return "FILTER ((!regex(str(?" + filter.attribute + "), '" + value + "', 'i'))) . ";
    } else if (filter.operator === "starts with") {
        value = "^" + filter.value;
        return "FILTER ((regex(str(?" + filter.attribute + "), '" + value + "', 'i'))) . ";
    } else if (filter.operator === "ends with") {
        value = filter.value + "$";
        return "FILTER ((regex(str(?" + filter.attribute + "), '" + value + "', 'i'))) . ";
    } else if (filter.operator === "regex") {
        value = filter.value;
        return "FILTER ((regex(str(?" + filter.attribute + "), '" + value + "', 'i'))) . ";
    } else {


        if (filter.dataType === "xsd:string") {
            // string but user used number operators !

            // TODO only int?
            var attr = "xsd:int(?" + filter.attribute + ")";

            return "FILTER (" + attr + " " + filter.operator + " " + filter.value + ") .";

        } else {

            // TODO hack - dateTime to date

            value = '"' + filter.value + '"';
            if (filter.dataType) {

                if (filter.dataType === "xsd:dateTime") {
                    value += "^^xsd:date";
                } else {
                    value += "^^" + filter.dataType;
                }

            }

            return "FILTER (?" + filter.attribute + " " + filter.operator + " " + value + ") .";
        }


    }

}

/**
 * returns selected attributes of given class
 * @param node
 * @returns {Array}
 */
function getSelectedAttributes(node) {

    var rtnAttributes = [];

    var attributes = classes[node].attributes;

    for (var i = 0; i < attributes.length; i++) {

        if (attributes[i].selected === true) {
            rtnAttributes.push(attributes[i]);
        }

    }

    return rtnAttributes;


}

/**
 * returns selected neighbours of given class (their object properties are selected)
 * @param node
 * @returns {Array}
 */
function getSelectedNeighbours(node) {
    var neighbours = [];

    for (var i = 0; i < graph.edges.length; i++) {
        if (graph.edges[i].source === node) {
            var objProperty = getObjectPropertyBySourceAndTarget(node, graph.edges[i].target, graph.edges[i].name);
            if (objProperty.selected === true) {
                neighbours.push(objProperty);
            }
        }
    }

    return neighbours;
}

/**
 * tries to find a starting node - a node which has outgoing relations and no incoming
 * can return also a node which has no incoming and outgoing relations
 * can also return the first selected node if no starting/isolated node is found
 * @param selectedNodes
 * @returns {*}
 */
function findStartingNode(selectedNodes) {

    for (var i = 0; i < selectedNodes.length; i++) {

        var value = selectedNodes[i];

        var areEdgesComing = areEdgesComingTo(value);
        var areEdgesOutcoming = areEdgesComingFrom(value);

        if (!areEdgesComing && areEdgesOutcoming) {
            // starting node
            return value;
        } else if (!areEdgesComing && !areEdgesOutcoming) {
            // isolated node?
            return value;
        }
    }

    return selectedNodes[0];

}

/**
 * checks if there are any relations coming from the node (selected object properties where source is this node and target other)
 * @param node
 * @returns {boolean}
 */
function areEdgesComingFrom(node) {
    for (var i = 0; i < graph.edges.length; i++) {

        if (graph.edges[i].source === node) {


            var objProperty = getObjectPropertyBySourceAndTarget(node, graph.edges[i].target, graph.edges[i].name);

            if (objProperty.selected === true) {
                return true;
            }
        }
    }

    return false;
}

/**
 * checks if there are any relations coming to this node (selected object properties where target is this node and source other)
 * @param node
 * @returns {boolean}
 */
function areEdgesComingTo(node) {

    for (var i = 0; i < graph.edges.length; i++) {

        if (graph.edges[i].target === node) {


            var objProperty = getObjectPropertyBySourceAndTarget(graph.edges[i].source, node, graph.edges[i].name);


            if (objProperty.selected === true) {
                return true;
            }
        }
    }

    return false;

}