/**
 * seeker.js
 * contains functions for graph search and traversal
 */

/**
 * checker for when we find a subclass
 * @type {boolean}
 */
var subClass = false;

/**
 * builds a path that has been traversed
 * @param target
 * @param path
 * @returns {*[]}
 */
function buildPath(target, path) {
    var result = [];

    while (path.has(target)) {
        var source = path.get(target);
        result.push({
            source: source,
            target: target
        });
        target = source;
    }

    return result.reverse();
}

/**
 * function to find a path between a source and target using BFS
 * returns null if no path can be found
 * otherwise returns first found path
 * @param source
 * @param target
 * @returns {null|*[]}
 */
function findPath(source, target) {

    // path to itself
    if (source === target) {
        return null;
    }

    var queue = [source];
    var visited = [];
    var path = new Map();

    while (queue.length > 0) {
        var start = queue.shift();

        // we found the target
        if (start === target) {
            return buildPath(start, path);
        }

        // get all neighbours
        var neighbours = getAllNeighbours(start);

        for (var next of neighbours) {

            // already visited?
            if (arrayIncludes(visited, next)) {
                continue;
            }

            // unless it is already in queue, push
            if (!arrayIncludes(queue, next)) {
                path.set(next, start);
                queue.push(next);
            }
        }
        visited.push(start);
    }

    return null;
}

/**
 * returns all neighbours of a node
 * @param node
 * @returns {Array}
 */
function getAllNeighbours(node) {
    var neighbours = [];

    for (var i = 0; i < graph.edges.length; i++) {
        if (graph.edges[i].source === node) {
            neighbours.push(graph.edges[i].target);
        }
    }

    return neighbours;
}

/**
 * updates graph with duplicate, meaning it duplicates edges of duplicated class to its duplicate
 * @param cls
 * @param duplicate
 */
function updateGraphWithDuplicate(cls, duplicate) {

    graph.nodes.push(duplicate);

    for (var i = 0; i < graph.edges.length; i++) {

        if (graph.edges[i].source === cls) {

            graph.edges.push({
                source: duplicate,
                target: graph.edges[i].target,
                name: graph.edges[i].name
            });
        }

        if (graph.edges[i].target === cls) {

            var property = getObjectPropertyBySourceAndTarget(graph.edges[i].source, graph.edges[i].target, graph.edges[i].name);

            var newProperty = createDuplicate(property);
            newProperty.target = duplicate;
            newProperty.duplicate = true;

            classes[graph.edges[i].source].objectProperties.push(newProperty);

            graph.edges.push({
                source: graph.edges[i].source,
                target: duplicate,
                name: graph.edges[i].name
            });
        }

    }


}

/**
 * updates graph in a sense that all nodes with id greater than length of classes are spliced
 * used when application deletes duplicates
 */
function updateGraph() {

    for (var i = graph.edges.length - 1; i >= 0; i--) {

        var edge = graph.edges[i];

        if (edge.source >= classes.length || edge.target >= classes.length) {
            graph.edges.splice(i, 1);
        }
    }


}

/**
 * function constructs a graph representation of classes array -> nodes and edges
 */
function constructGraph() {

    graph.nodes = [];
    graph.edges = [];

    for (var i = 0; i < classes.length; i++) {

        graph.nodes.push(i);

        var objectProperties = classes[i].objectProperties;

        if (objectProperties.length > 0) {

            for (var j = objectProperties.length - 1; j >= 0; j--) {

                var objectProperty = objectProperties[j];

                var target = getTargetFromProperty(objectProperty);

                // unknown target class
                if (target === -1) {

                    // TODO ciselniky
                    // here is a problem with classes that are not present in the dataset like "ciselniky"
                    // if left in object properties, target is unknown and it would not work
                    // now it is done so these properties are switched to attributes
                    // other solution would be to try and load the target class from ontologies, since they are not in dataset

                    objectProperties.splice(j, 1);

                    classes[i].attributes.push(objectProperty);
                    classes[i].attributes.sort(compareCount);

                    // the target is a subclass
                } else if (subClass === true) {

                    subClass = false;

                    // look for all subclasses of this type
                    var classesWithThemSameSubClass = getAllClassesOfSubclass(getRange(objectProperty));

                    if (classesWithThemSameSubClass.length > 1) {
                        // more than 1

                        // the first one is already in target
                        objectProperty.target = target;

                        graph.edges.push({source: i, target: target, name: objectProperty.localName});


                        // add the rest in
                        for (var k = 1; k < classesWithThemSameSubClass.length; k++) {

                            var duplicate = createDuplicate(objectProperty);

                            duplicate.target = classesWithThemSameSubClass[k];

                            graph.edges.push({source: i, target: duplicate.target, name: objectProperty.localName});

                            objectProperties.push(duplicate);

                            objectProperties.sort(compareCount);

                        }


                    } else {
                        // normal then
                        objectProperty.target = target;

                        graph.edges.push({source: i, target: target, name: objectProperty.localName});
                    }

                    // not unknown and not subclass
                } else {
                    objectProperty.target = target;

                    graph.edges.push({source: i, target: target, name: objectProperty.localName});

                }
            }
        }
    }

}