/**
 * visualization.js
 * d3 visualization of generated SPARQL query
 */

var w = 900;
var h = 600;
var linkDistance = 200;

var colors = d3.scale.category10();

var dataset = {
    nodes: [],
    edges: []
};


var svg = d3.select("svg").attr({"width": w, "height": h});

var edges = svg.selectAll("line");
var nodes = svg.selectAll("circle");
var nodelabels = svg.selectAll(".nodelabel");
var edgepaths = svg.selectAll(".edgepath");
var edgelabels = svg.selectAll(".edgelabel");

var force = d3.layout.force()
    .size([w, h])
    .linkDistance([linkDistance])
    .charge([-500])
    .theta(0.1)
    .gravity(0.05)
    .on("tick", ticked);


restart();

/**
 * redraws the visualization
 */
function restart() {


    svg.selectAll("*").remove();


    svg.append('defs').append('marker')
        .attr({
            'id': 'arrowhead',
            'viewBox': '-0 -5 10 10',
            'refX': 25,
            'refY': 0,
            'orient': 'auto',
            'markerWidth': 10,
            'markerHeight': 10,
            'xoverflow': 'visible'
        })
        .append('svg:path')
        .attr('d', 'M 0,-5 L 10 ,0 L 0,5')
        .attr('fill', '#ccc')
        .attr('stroke', '#ccc');

    edges = svg.selectAll("line");
    nodes = svg.selectAll("circle");
    nodelabels = svg.selectAll(".nodelabel");
    edgepaths = svg.selectAll(".edgepath");
    edgelabels = svg.selectAll(".edgelabel");

    edges = edges.data(dataset.edges);

    edges.exit().remove();

    edges = edges.enter()
        .append("line")
        .attr("id", function (d, i) {
            return 'edge' + i
        })
        .attr('marker-end', 'url(#arrowhead)')
        .style("stroke", "#ccc")
        .style("pointer-events", "none");

    nodes = nodes.data(dataset.nodes);

    nodes.exit().remove();

    nodes = nodes
        .enter()
        .append("circle")
        .attr({
            "r": function (d) {
                return (d.id === currentClass) ? 18 : 15;
            }
        })
        .style('fill', function (d) {
            return colors(d.id)
        })
        .call(force.drag).on("click", function (d) {
            var selectedNode = d.id;

            if (currentClass >= 0) {
                if (currentAttribute !== -1 && currentClass !== -1) {
                    classes[currentClass].attributes[currentAttribute].clicked = false;
                }
            }


            currentClass = selectedNode;

            currentAttribute = -1;

            showInfo();
            createObjectPropertiesSelection();
            createObjectPropertiesTable();
            createAttributesSelection();
            createAttributeTable();

            clearFiltersAndAggregations();

            restart();
        });
    nodelabels = nodelabels.data(dataset.nodes);

    nodelabels.exit().remove();

    nodelabels = nodelabels
        .enter()
        .append("text")
        .attr({
            "x": function (d) {
                return d.x;
            },
            "y": function (d) {
                return d.y;
            },
            "class": "nodelabel",
            "stroke": "black"
        })
        .text(function (d) {
            return d.name;
        });

    edgepaths = edgepaths.data(dataset.edges);

    edgepaths.exit().remove();

    edgepaths = edgepaths
        .enter()
        .append('path')
        .attr({
            'd': function (d) {
                return 'M ' + d.source.x + ' ' + d.source.y + ' L ' + d.target.x + ' ' + d.target.y
            },
            'class': 'edgepath',
            'fill-opacity': 0,
            'stroke-opacity': 0,
            'fill': 'blue',
            'stroke': 'red',
            'id': function (d, i) {
                return 'edgepath' + i
            }
        })
        .style("pointer-events", "none");

    edgelabels = edgelabels.data(dataset.edges);

    edgelabels.exit().remove();

    edgelabels = edgelabels
        .enter()
        .append('text')
        .style("pointer-events", "none")
        .attr({
            'class': 'edgelabel',
            'id': function (d, i) {
                return 'edgelabel' + i
            },
            'dx': 60,
            'dy': 0,
            'font-size': 10,
            'fill': '#aaa'
        });

    edgelabels.append('textPath')
        .attr('xlink:href', function (d, i) {
            return '#edgepath' + i
        })
        .style("pointer-events", "none")
        .text(function (d) {
            return d.name
        });


    force.nodes(dataset.nodes);
    force.links(dataset.edges);
    force.start();


}

/**
 * called on each tick - the update of a visualization
 */
function ticked() {

    edges.attr({
        "x1": function (d) {
            return d.source.x;
        },
        "y1": function (d) {
            return d.source.y;
        },
        "x2": function (d) {
            return d.target.x;
        },
        "y2": function (d) {
            return d.target.y;
        }
    });

    nodes.attr({
        "cx": function (d) {
            return d.x;
        },
        "cy": function (d) {
            return d.y;
        }
    });

    nodelabels.attr("x", function (d) {
        return d.x;
    })
        .attr("y", function (d) {
            return d.y;
        });

    edgepaths.attr('d', function (d) {
        return 'M ' + d.source.x + ' ' + d.source.y + ' L ' + d.target.x + ' ' + d.target.y;
    });

    edgelabels.attr('transform', function (d) {
        if (d.target.x < d.source.x) {
            bbox = this.getBBox();
            rx = bbox.x + bbox.width / 2;
            ry = bbox.y + bbox.height / 2;
            return 'rotate(180 ' + rx + ' ' + ry + ')';
        } else {
            return 'rotate(0)';
        }
    });
}

/**
 *
 * @param newNode
 */
function addNode(newNode) {

    dataset.nodes.push(newNode);

    restart();
}

/**
 *
 * @param newEdge
 */
function addEdge(newEdge) {
    dataset.edges.push(newEdge);

    restart();
}

/**
 * removes an edge from graph
 * @param source
 * @param target
 */
function removeEdge(source, target, name) {

    for (var j = dataset.edges.length - 1; j >= 0; j--) {

        if (dataset.edges[j].source.id === source && dataset.edges[j].target.id === target && dataset.edges[j].edgeName === name) {

            dataset.edges.splice(j, 1);
            break;

        }

    }


}

/**
 * removes node from graph
 * @param id
 */
function removeNode(id) {

    for (var i = dataset.edges.length - 1; i >= 0; i--) {

        if (dataset.edges[i].source.id === id || dataset.edges[i].target.id === id) {

            var obj = getObjectPropertyBySourceAndTarget(dataset.edges[i].source.id, dataset.edges[i].target.id, dataset.edges[i].edgeName);

            obj.selected = false;

            dataset.edges.splice(i, 1);

        }
    }

    for (var j = dataset.nodes.length - 1; j >= 0; j--) {

        if (dataset.nodes[j].id === id) {
            dataset.nodes.splice(j, 1);
            break;
        }
    }
}

/**
 * updates ids after removal of nodes/edges
 */
function updateIds() {

    for (var i = 0; i < dataset.nodes.length; i++) {


        var node = dataset.nodes[i];

        var cls = classes[node.id];
        cls.d3Id = node.index;

    }


}
